/*
 * Title:        bits/custom_t.hpp
 * Description:  C++ header-only wrapper for C-ABI data_t type
 * $Copyright (c) 2015-2016 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: custom_t.hpp 320837 2017-11-30 21:07:30Z bsp $
 */

/* This file:
 * Contains the C++ implementation of custom_t
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif


/**
 * A wrapper type for holding arbitrary objects inside a data_t.
 * It behaves like a std::unique_ptr with an explict copy method and will
 * delete the contents when destroyed unless they are moved out first.
 *
 * The wrapped class must have a nothrow destructor and a nothrow copy constructor.
 *
 * You must ensure that you extract your object with the same type that was inserted,
 * or the behaviour is undefined. If you have a hierarchy of objects and you need
 * to determine the type at runtime then you must extract as a custom_t<BaseClass> and
 * then dynamic_cast or call virtual methods to determine the type. This means you
 * must have a virtual hierarchy of objects.
 *
 * This can be used with get<> or get_custom<> to retrieve the contents of a data_t:
 *
 * get<custom_t<MyClass> >(data);
 * get_custom<MyClass>(data);
 *
 * however visitors will still visit a sag_underlying_custom_t& which can
 * be cast to a custom_t<MyClass>&.
 */
template<typename T>
class custom_t: public sag_underlying_custom_t
{
	typedef custom_t this_t;
public:
	/** Create an empty custom_t */
	custom_t();
	/** Create a custom_t which owns the given payload. */
	custom_t(T *payload);
	/** deletes the underlying pointer. */
	~custom_t();
	/** Return the wrapped pointer as the correct type. */
	T *get() const { return this->data ? reinterpret_cast<T*>(this->data->custom) : 0; }
	/** Move constructor. */
	custom_t(custom_t &&other);
	/** Move assignment, destroying the current contents. */
	custom_t &operator=(custom_t &&other);
	/** Swap with another custom_t of the same type. */
	void swap(custom_t &&other);
	/** Return a copy of the underlying pointer. */
	custom_t copy() const;
	/** Return the wrapped pointer as the correct type. */
	T &operator*() const { return *get(); }
	/** Return the wrapped pointer as the correct type. */
	T *operator->() const { return get(); }
private:
	/** @internal static method which can be passed over the C boundry to free the underlying pointer */
	static void deleter(void *d) 
	{
		if (d) {
			try {
				delete reinterpret_cast<T*>(d); 
			} catch (...) {
				// enforce nothrow behaviour
				abort();
			}
		}
	}
	/** @internal static method which can be passed over the C boundry to copy the underlying pointer */
	static void *copier(void *d)
	{
		try {
			return d ? new T(*reinterpret_cast<T*>(d)) : 0; 
		} catch (...) {
			// enforce nothrow behaviour
			abort();
		}
	}
	/** @internal not copiable */
	custom_t(const custom_t &other) { abort(); }
	/** @internal not copiable */
	custom_t &operator=(const custom_t &other) { abort(); return *this; }
};

